# Maze Generator with Loop-erased Random Walks
#
# Copyright 2021 William Kearney
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using SparseArrays, Sockets

struct GridTree
    A
    N
    function GridTree(N)
        A = spzeros(Bool,N*N,N*N)
        A[1,1] = 1
        new(A,N)
    end
end

function intree(p,A::GridTree)
    A,N = A.A,A.N
    i,j = p
    any(A[:,(j-1)*N + i])
end

function add_path!(A::GridTree,path)
    A,N = A.A,A.N
    for k in 2:length(path)
        u = path[k]
        i,j = u
        v = path[k-1]
        l,m = v
        
        A[(j-1)*N + i,(m-1)*N+l] = 1
        A[(m-1)*N + l,(j-1)*N+i] = 1
    end
end

function neighbors(u,N)
    i,j = u
    n = (i,j+1)
    e = (i+1,j)
    s = (i,j-1)
    w = (i-1,j)
    if i == 1
        if j == 1
            return [n;e]
        elseif j == N
            return [e;s]
        else
            return [n;e;s]
        end
    elseif i == N
        if j == 1
            return [n;w]
        elseif j == N
            return [s;w]
        else
            return [n;s;w]
        end
    else
        if j == 1
            return [n;e;w]
        elseif j == N
            return [e;s;w]
        else
            return [n;e;s;w]
        end
    end        
end

function rw_step(u,N)
    ns = neighbors(u,N)
    rand(ns)    
end

function rw(u0,T,N)
    us = [(u0[1],u0[2])]
    while !intree(us[end],T)
        push!(us,rw_step(us[end],N))
    end
    us
end

function erase_loops(γ)
    N = length(γ)
    is = [1]
    while is[end] < N        
        push!(is,findlast(x->x == γ[is[end]],γ)+1)
    end
    γ[is]
end

function lera(u0,T,N)
    γ = rw(u0,T,N)
    erase_loops(γ)
end

function find_first_v(T::GridTree,vs)
    for v in vs
        if !intree(v,T)
            return v
        end
    end        
end

function ust(u0,N)
    vs = vec([(i,j) for i in N:-1:1, j in N:-1:1])
    T = GridTree(N)
    v = find_first_v(T,vs)
    while !isnothing(v)
        γ = lera(v,T,N)
        add_path!(T,γ)
        v = find_first_v(T,vs)
    end
    T
end

function get_directions(k,t::GridTree)
    N = t.N
    neighbors = t.A[k,:].nzind .-k
    [in(N,neighbors),  # North
     in(1,neighbors),  # East
     in(-N,neighbors), # South
     in(-1,neighbors)] # West
end

function render_line(j,t)
    N = t.N
    if j > N
        s = "_"^(2N-1)*" _"
    else
        if (j ==1) # Make the entrance
            s = "_"
        else
            s = "|"
        end
        for i in 1:N
            directions = get_directions(i+(j-1)*N,t)
            if directions[2] # East
                if directions[3] # South
                    s *= " _"
                else
                    s *= "__"
                end
            else
                if directions[3]
                    s *= " |"
                else
                    s *= "_|"
                end
            end
        end
    end
    s
end

function Base.show(io::IO,x::GridTree)
    N = x.N
    for i in N+1:-1:1
        println(io,render_line(i,x))
    end
end

function start_mazeserver(domain_socket)
    server = listen(domain_socket)
    chmod(domain_socket,0o777)
    begin
        #println("Maze server is listening...")
        while isopen(server)
            sock = accept(server)
            #println("Socket opened")
	    Nheader = parse(Int,readuntil(sock,":",keep=false))
	    #println("Header length: ",Nheader)
	    header = transcode(String,read(sock,Nheader+1))
            # Gemini Response Header
            print(sock,"20 text/gemini;charset=utf-8\r\n")
            #
            print(sock,"# A randomly generated maze\r\n")
            print(sock,"Make your way from the bottom left to the top right\r\n")
            # Maze	    
            print(sock,"```\r\n",ust((1,1),16),"```\r\n\r\n")
	    close(sock)
        end
    end
end